class Logger:
    """
    A simple logger class for printing log messages based on the specified log level.

    Attributes:
    logLevel (str): The log level to determine which messages to print ('verbose' or 'off').

    Methods:
    __init__(self, logLevel='verbose'): Initializes the Logger object with the specified log level.
    log(self, message): Logs the given message based on the log level.
    """

    def __init__(self, logLevel='verbose'):
        """
        Initialize the Logger object.

        Parameters:
        logLevel (str): The log level to determine which messages to print ('verbose' or 'off').
                        Default is 'verbose'.
        """
        self.logLevel = logLevel

    def log(self, message):
        """
        Log the given message based on the log level.

        Parameters:
        message (str): The message to be logged.
        """
        if self.logLevel == 'verbose':
            print(message)



import torch
from torch.autograd import Variable

from torch.utils.data import  DataLoader

from routy import Comparison
from routy.distanceUtilities.distanceUtilities import DistanceType
from routy.modelsForComparison.tsp import CombinatorialRL, reward, TSPDataset
import matplotlib

from routy.plotUtilities.plotRoute import plot_multiple_routes_for_same_coordinates

matplotlib.use('Agg')



if __name__ == '__main__':

        num_nodes = 10
        seed = 12345
        valid_size = 100
        embedding_size = 128
        n_glimpses = 1
        hidden_size = 128
        tanh_exploration = 10
        use_tanh = False

        comparison = Comparison()
        model = CombinatorialRL(
            embedding_size,
            hidden_size,
            num_nodes,
            n_glimpses,
            tanh_exploration,
            use_tanh,
            reward,
            embedding="Conv",
            attention="Dot")

        path = "tsp_20_dot_attention.pt"
        model.load_state_dict(torch.load(path))

        train_size = 15
        val_size = 1000
        batch_size = 100
        num_nodes = 20
        epochs = 20

        train_20_dataset = TSPDataset(num_nodes, train_size)
        loader = DataLoader(train_20_dataset,
                            batch_size=batch_size,
                        shuffle=True, num_workers=1)
        tsp_dot_attention_model20_path = "tsp_20_dot_attention.pt"

        tsp_dot_attention_model20_trained = CombinatorialRL(
            embedding_size,
            hidden_size,
            20,
            n_glimpses,
            tanh_exploration,
            use_tanh,
            reward, embedding="Conv",
            attention="Dot")
        tsp_dot_attention_model20_trained.load_state_dict(torch.load(tsp_dot_attention_model20_path))
        tsp_dot_attention_model20_trained.eval()

        # We compare with an untrained model
        tsp_20_model_untrained = CombinatorialRL(
            embedding_size,
            hidden_size,
            20,
            n_glimpses,
            tanh_exploration,
            use_tanh,
            reward,
            embedding="Conv",
            attention="Dot")
        tsp_20_model_untrained.eval()
        test_size = 100
        test_20_dataset = TSPDataset(20, test_size)
        test_loader_20nodes = DataLoader(test_20_dataset, batch_size=100, shuffle=True, num_workers=1)

        test_5_dataset = TSPDataset(5, test_size)
        test_loader = DataLoader(test_5_dataset, batch_size=2, shuffle=True, num_workers=1)

        tsp_dot_attention_model20_path = "tsp_20_dot_attention.pt"

        tsp_dot_attention_model20_trained = CombinatorialRL(
            embedding_size,
            hidden_size,
            20,
            n_glimpses,
            tanh_exploration,
            use_tanh,
            reward, embedding="Conv",
            attention="Dot")
        tsp_dot_attention_model20_trained.load_state_dict(torch.load(tsp_dot_attention_model20_path))
        tsp_dot_attention_model20_trained.eval()

        # We compare with an untrained model
        tsp_20_model_untrained = CombinatorialRL(
            embedding_size,
            hidden_size,
            20,
            n_glimpses,
            tanh_exploration,
            use_tanh,
            reward, embedding="Conv",
            attention="Dot")
        tsp_20_model_untrained.eval()


        # for batch_id, sample_batch in enumerate(test_loader):
        #     inputs = Variable(sample_batch)
        #     R_1, probs_1, actions_1, actions_idxs_trained1 = tsp_dot_attention_model20_trained(inputs)
        #
        #     R, probs, actions, actions_idxs = tsp_20_model_untrained(inputs)
        #
        #     print(f'{actions_idxs}')
        #     tensors = actions_idxs
        #
        #     first_column_row = torch.cat([tensor[0].unsqueeze(0) for tensor in tensors])
        #     sec_column_row = torch.cat([tensor[1].unsqueeze(0) for tensor in tensors])
        #
        #
        #     print(f'{first_column_row}')
        #     print(f'{sec_column_row}')
        #     assert inputs.size(1) == 2
        #     assert actions_1[0][0].size(0) == 2
        #
        #     "inputs: [test_size, 2, seq_len]"
        #     "actions_1: [seq_len, test_size, 2]"
        #
        #     route = first_column_row.tolist()
        #     multiple_routes = [first_column_row.tolist(), sec_column_row.tolist()]
        #     coordinates = inputs[0].transpose(0,1).tolist()
        #
        #
        #     comparison.plot_route(coordinates,route)
        #     comparison.plot_routes(coordinates, multiple_routes)

        routes1 = []
        routes2 = []
        coordinates = []
        for batch_id, sample_batch in enumerate(test_loader):
            inputs = Variable(sample_batch)
            R_1, probs_1, actions_1, actions_idxs_trained1 = tsp_dot_attention_model20_trained(inputs)

            R, probs, actions, actions_idxs = tsp_20_model_untrained(inputs)

            print(f'{actions_idxs}')
            tensors = actions_idxs

            first_column_row = torch.cat([tensor[0].unsqueeze(0) for tensor in tensors])
            sec_column_row = torch.cat([tensor[1].unsqueeze(0) for tensor in tensors])


            print(f'{first_column_row}')
            print(f'{sec_column_row}')
            assert inputs.size(1) == 2
            assert actions_1[0][0].size(0) == 2

            "inputs: [test_size, 2, seq_len]"
            "actions_1: [seq_len, test_size, 2]"

            # route = first_column_row.tolist()
            # multiple_routes = [first_column_row.tolist(), sec_column_row.tolist()]
            # coordinates = inputs[0].transpose(0,1).tolist()

            coords = inputs[0].transpose(0, 1).tolist()
            coordinates.append(coords)
            routes1.append(first_column_row.tolist())
            routes2.append(sec_column_row.tolist())


        comparison.compare_model_distances(coordinates, routes1, routes2, DistanceType.EUCLIDEAN)
        comparison.compare_model_distances(coordinates, routes1, routes2, DistanceType.MANHATTAN)
        comparison.compare_model_distances(coordinates, routes1, routes2, DistanceType.CHEBYSHEV)


from routy.Logger import Logger
from routy.distanceUtilities.distanceUtilities import calculate_distance, DistanceType
from routy.plotUtilities.plotRoute import plot_multiple_routes_for_same_coordinates, plot_route


class Comparison:
    """
    Instantiate a comparison operation. This class is the starting point for comparing your
    routing models.
    :param None
    """

    def __init__(self):

        self.logger = Logger()

    def plot_route(self, coordinates, routes, filename="route"):
        '''
        Plot a single route based on coordinates and routes.

        :param coordinates: List of coordinates [(x, y)] for each point on the route.
        :type coordinates: list of tuples
        :param routes: List of indices representing the route.
        :type routes: list
        :param filename: Name of the output plot file (default: "route.png").
        :type filename: str
         :Examples:
        plot_route([(0, 0), (1, 1), (2, 2)], [0, 1, 2], "example_route.png")
        '''

        seq_len = len(routes)
        assert seq_len == len(coordinates)
        assert type(coordinates[0]) is tuple or len(coordinates[1]) ==2

        plot_route(coordinates, routes, f'{filename}.png')

    def plot_routes(self,coordinates, routes, filename="routes"):
        '''
        Plot multiple routes for the same set of coordinates.

        :param coordinates: List of coordinates [(x, y)] for each point on the routes.
        :type coordinates: list of tuples
        :param routes: List of lists, each representing a route with indices.
        :type routes: list of lists
        :param filename: Name of the output plot file (default: "routes.png").
        :type filename: str

         :Examples:
        plot_routes([(0, 0), (1, 1), (2, 2)], [[0, 1, 2], [2, 1, 0]], "example_routes.png")
        '''
        seq_len = len(routes[0])
        assert seq_len == len(coordinates)

        assert type(coordinates[0]) is tuple or  len(coordinates[0]) == 2
        assert type(routes) is list
        plot_multiple_routes_for_same_coordinates(coordinates,
                                                  routes,
                                                  filename)
    def compare_model_distances(self, coordinates, model1Routes, model2Routes, distance_type=DistanceType.EUCLIDEAN):
        """
        Calculate the percentage of routes where model1 suggests a shorter route compared to model2.

        Parameters:
        model1_routes (list): List of routes suggested by model1.
        model2_routes (list): List of routes suggested by model2.
        Returns:
        float: Percentage of routes where model1 suggests a shorter route compared to model2.
        """
        # assert they are a list of lists
        assert isinstance(coordinates, list)  and all(isinstance(coordinates, list) for item in coordinates)
        assert isinstance(model1Routes, list) and all(isinstance(model1Routes, list) for item in model1Routes)
        assert isinstance(model2Routes, list) and all(isinstance(model2Routes, list) for item in model2Routes)

        total_routes = len(coordinates)
        assert total_routes == len(model1Routes)
        assert total_routes == len(model2Routes)


        shorter_routes_count = 0

        for i in range(total_routes):
            coords = coordinates[i]
            route1 = model1Routes[i]
            route2 = model2Routes[i]
            dist1 = calculate_distance(coords, route1, distance_type)
            dist2 = calculate_distance(coords, route2, distance_type)

            if dist1 < dist2:
                shorter_routes_count += 1


        percentage_shorter_routes = (shorter_routes_count / total_routes) * 100

        self.logger.log(f'Using {distance_type} distance')
        self.logger.log(f'Model 1 has {percentage_shorter_routes:.2f}% shorter routes from model 2, for {total_routes} total routes')
        return percentage_shorter_routes
